@@ -1,5 +1,5 @@
 /*
- *	$Id: lspci.c,v 1.32 2000/01/13 22:50:13 mj Exp $
+ *	$Id: lspci.c,v 1.38 2000/05/20 14:36:02 mj Exp $
  *
  *	Linux PCI Utilities -- List All PCI Devices
  *
@@ -74,6 +74,16 @@ static struct pci_access *pacc;
 #define IO_FORMAT "%04lx"
 #endif
 
+/*
+ *  If we aren't being compiled by GCC, use malloc() instead of alloca().
+ *  This increases our memory footprint, but only slightly since we don't
+ *  use alloca() much.
+ */
+
+#ifndef __GNUC__
+#define alloca malloc
+#endif
+
 /* Our view of the PCI bus */
 
 struct device {
@@ -101,7 +111,7 @@ scan_device(struct pci_dev *p)
   if (how_much < 128 && (d->config[PCI_HEADER_TYPE] & 0x7f) == PCI_HEADER_TYPE_CARDBUS)
     {
       /* For cardbus bridges, we need to fetch 64 bytes more to get the full standard header... */
-      if (!pci_read_block(p, 0, d->config+64, 64))
+      if (!pci_read_block(p, 64, d->config+64, 64))
 	die("Unable to read cardbus bridge extension data.");
       how_much = 128;
     }
@@ -365,32 +375,63 @@ show_bases(struct device *d, int cnt)
 static void
 show_pm(struct device *d, int where, int cap)
 {
-  int t;
+  int t, b;
+  static int pm_aux_current[8] = { 0, 55, 100, 160, 220, 270, 320, 375 };
 
   printf("Power Management version %d\n", cap & PCI_PM_CAP_VER_MASK);
   if (verbose < 2)
     return;
-  printf("\t\tFlags: PMEClk%c AuxPwr%c DSI%c D1%c D2%c PME%c\n",
+  printf("\t\tFlags: PMEClk%c DSI%c D1%c D2%c AuxCurrent=%dmA PME(D0%c,D1%c,D2%c,D3hot%c,D3cold%c)\n",
 	 FLAG(cap, PCI_PM_CAP_PME_CLOCK),
-	 FLAG(cap, PCI_PM_CAP_AUX_POWER),
 	 FLAG(cap, PCI_PM_CAP_DSI),
 	 FLAG(cap, PCI_PM_CAP_D1),
 	 FLAG(cap, PCI_PM_CAP_D2),
-	 FLAG(cap, PCI_PM_CAP_PME));
+	 pm_aux_current[(cap >> 6) & 7],
+	 FLAG(cap, PCI_PM_CAP_PME_D0),
+	 FLAG(cap, PCI_PM_CAP_PME_D1),
+	 FLAG(cap, PCI_PM_CAP_PME_D2),
+	 FLAG(cap, PCI_PM_CAP_PME_D3_HOT),
+	 FLAG(cap, PCI_PM_CAP_PME_D3_COLD));
   config_fetch(d, where + PCI_PM_CTRL, PCI_PM_SIZEOF - PCI_PM_CTRL);
   t = get_conf_word(d, where + PCI_PM_CTRL);
-  printf("\t\tStatus: D%d PME-Enable%c DSel=%x DScale=%x PME%c\n",
+  printf("\t\tStatus: D%d PME-Enable%c DSel=%d DScale=%d PME%c\n",
 	 t & PCI_PM_CTRL_STATE_MASK,
 	 FLAG(t, PCI_PM_CTRL_PME_ENABLE),
 	 (t & PCI_PM_CTRL_DATA_SEL_MASK) >> 9,
 	 (t & PCI_PM_CTRL_DATA_SCALE_MASK) >> 13,
 	 FLAG(t, PCI_PM_CTRL_PME_STATUS));
+  b = get_conf_byte(d, where + PCI_PM_PPB_EXTENSIONS);
+  if (b)
+    printf("\t\tBridge: PM%c B3%c\n",
+	   FLAG(t, PCI_PM_BPCC_ENABLE),
+	   FLAG(~t, PCI_PM_PPB_B2_B3));
+}
+
+static void
+format_agp_rate(int rate, char *buf)
+{
+  char *c = buf;
+  int i;
+
+  for(i=0; i<2; i++)
+    if (rate & (1 << i))
+      {
+	if (c != buf)
+	  *c++ = ',';
+	*c++ = 'x';
+	*c++ = '0' + (1 << i);
+      }
+  if (c != buf)
+    *c = 0;
+  else
+    strcpy(buf, "<none>");
 }
 
 static void
 show_agp(struct device *d, int where, int cap)
 {
   u32 t;
+  char rate[8];
 
   t = cap & 0xff;
   printf("AGP version %x.%x\n", cap/16, cap%16);
@@ -398,24 +439,22 @@ show_agp(struct device *d, int where, int cap)
     return;
   config_fetch(d, where + PCI_AGP_STATUS, PCI_AGP_SIZEOF - PCI_AGP_STATUS);
   t = get_conf_long(d, where + PCI_AGP_STATUS);
-  printf("\t\tStatus: RQ=%d SBA%c 64bit%c FW%c Rate=%s%s%s\n",
+  format_agp_rate(t & 7, rate);
+  printf("\t\tStatus: RQ=%d SBA%c 64bit%c FW%c Rate=%s\n",
 	 (t & PCI_AGP_STATUS_RQ_MASK) >> 24U,
 	 FLAG(t, PCI_AGP_STATUS_SBA),
 	 FLAG(t, PCI_AGP_STATUS_64BIT),
 	 FLAG(t, PCI_AGP_STATUS_FW),
-	 (t & PCI_AGP_STATUS_RATE4) ? "4" : "",
-	 (t & PCI_AGP_STATUS_RATE2) ? "2" : "",
-	 (t & PCI_AGP_STATUS_RATE1) ? "1" : "");
+	 rate);
   t = get_conf_long(d, where + PCI_AGP_COMMAND);
-  printf("\t\tCommand: RQ=%d SBA%c AGP%c 64bit%c FW%c Rate=%s%s%s\n",
+  format_agp_rate(t & 7, rate);
+  printf("\t\tCommand: RQ=%d SBA%c AGP%c 64bit%c FW%c Rate=%s\n",
 	 (t & PCI_AGP_COMMAND_RQ_MASK) >> 24U,
 	 FLAG(t, PCI_AGP_COMMAND_SBA),
 	 FLAG(t, PCI_AGP_COMMAND_AGP),
 	 FLAG(t, PCI_AGP_COMMAND_64BIT),
 	 FLAG(t, PCI_AGP_COMMAND_FW),
-	 (t & PCI_AGP_COMMAND_RATE4) ? "4" : "",
-	 (t & PCI_AGP_COMMAND_RATE2) ? "2" : "",
-	 (t & PCI_AGP_COMMAND_RATE1) ? "1" : "");
+	 rate);
 }
 
 static void
@@ -644,7 +683,7 @@ show_htype2(struct device *d)
       u32 base = get_conf_long(d, PCI_CB_MEMORY_BASE_0 + p);
       u32 limit = get_conf_long(d, PCI_CB_MEMORY_LIMIT_0 + p);
       if (limit > base || verb)
-	printf("Memory window %d: %08x-%08x%s%s\n", i, base, limit,
+	printf("\tMemory window %d: %08x-%08x%s%s\n", i, base, limit,
 	       (cmd & PCI_COMMAND_MEMORY) ? "" : " [disabled]",
 	       (brc & (PCI_CB_BRIDGE_CTL_PREFETCH_MEM0 << i)) ? " (prefetchable)" : "");
     }
@@ -704,11 +743,7 @@ show_verbose(struct device *d)
     {
     case PCI_HEADER_TYPE_NORMAL:
       if (class == PCI_CLASS_BRIDGE_PCI)
-	{
-	badhdr:
-	  printf("\t!!! Header type %02x doesn't match class code %04x\n", htype, class);
-	  return;
-	}
+	printf("\t!!! Invalid class %04x for header type %02x\n", class, htype);
       max_lat = get_conf_byte(d, PCI_MAX_LAT);
       min_gnt = get_conf_byte(d, PCI_MIN_GNT);
       subsys_v = get_conf_word(d, PCI_SUBSYSTEM_VENDOR_ID);
@@ -716,13 +751,13 @@ show_verbose(struct device *d)
       break;
     case PCI_HEADER_TYPE_BRIDGE:
       if (class != PCI_CLASS_BRIDGE_PCI)
-	goto badhdr;
+	printf("\t!!! Invalid class %04x for header type %02x\n", class, htype);
       irq = int_pin = min_gnt = max_lat = 0;
       subsys_v = subsys_d = 0;
       break;
     case PCI_HEADER_TYPE_CARDBUS:
       if ((class >> 8) != PCI_BASE_CLASS_BRIDGE)
-	goto badhdr;
+	printf("\t!!! Invalid class %04x for header type %02x\n", class, htype);
       min_gnt = max_lat = 0;
       subsys_v = get_conf_word(d, PCI_CB_SUBSYSTEM_VENDOR_ID);
       subsys_d = get_conf_word(d, PCI_CB_SUBSYSTEM_ID);
@@ -767,12 +802,18 @@ show_verbose(struct device *d)
 	     FLAG(status, PCI_STATUS_DETECTED_PARITY));
       if (cmd & PCI_COMMAND_MASTER)
 	{
-	  printf("\tLatency: ");
-	  if (min_gnt)
-	    printf("%d min, ", min_gnt);
-	  if (max_lat)
-	    printf("%d max, ", max_lat);
-	  printf("%d set", latency);
+	  printf("\tLatency: %d", latency);
+	  if (min_gnt || max_lat)
+	    {
+	      printf(" (");
+	      if (min_gnt)
+		printf("%dns min", min_gnt*250);
+	      if (min_gnt && max_lat)
+		printf(", ");
+	      if (max_lat)
+		printf("%dns max", max_lat*250);
+	      putchar(')');
+	    }
 	  if (cache_line)
 	    printf(", cache line size %02x", cache_line);
 	  putchar('\n');
@@ -987,7 +1028,10 @@ insert_dev(struct device *d, struct bridge *b)
       struct bridge *c;
       for(c=b->child; c; c=c->next)
 	if (c->secondary <= p->bus && p->bus <= c->subordinate)
-	  return insert_dev(d, c);
+          {
+            insert_dev(d, c);
+            return;
+          }
       bus = new_bus(b, p->bus);
     }
   /* Simple insertion at the end _does_ guarantee the correct order as the
